"""Additional data types"""

import numpy as np

import nada_dsl as dsl

from nada_dsl import (
    Input,
    Party,
    UnsignedInteger,
    Integer,
    NadaType,
    SecretInteger,
    SecretUnsignedInteger,
    PublicInteger,
    PublicUnsignedInteger,
)


from typing import Union


_Number = Union[float, int, np.floating]
_NadaInteger = Union[
    Integer,
    PublicInteger,
]
_NadaLike = Union[int, "Rational", "SecretRational", NadaType]

_NadaRational = Union["Rational", "SecretRational"]

_NadaType = Union[
    Integer,
    PublicInteger,
    PublicUnsignedInteger,
    SecretInteger,
    SecretUnsignedInteger,
    UnsignedInteger,
]


class SecretBoolean(dsl.SecretBoolean):

    def __init__(self, value):
        super().__init__(value.inner)

    def if_else(
        self: dsl.SecretBoolean,
        arg_0: _NadaType | "SecretRational" | "Rational",
        arg_1: _NadaType | "SecretRational" | "Rational",
    ) -> Union[SecretInteger, SecretUnsignedInteger]:
        first_arg = arg_0
        second_arg = arg_1
        if isinstance(arg_0, (SecretRational, Rational)) and isinstance(
            arg_1, (SecretRational, Rational)
        ):
            # Both are SecretRational or Rational objects
            if arg_0.log_scale != arg_1.log_scale:
                raise ValueError("Cannot output values with different scales.")
            first_arg = arg_0.value
            second_arg = arg_1.value
        elif isinstance(arg_0, (Rational, SecretRational)) or isinstance(
            arg_1, (Rational, SecretRational)
        ):
            # Both are SecretRational or Rational objects
            raise TypeError(f"Invalid operation: {self}.IfElse({arg_0}, {arg_1})")

        result = super().if_else(first_arg, second_arg)

        if isinstance(arg_0, (SecretRational, Rational)):
            # If we have a SecretBoolean, the return type will be SecretInteger, thus promoted to SecretRational
            return SecretRational.from_parts(result, arg_0.log_scale)
        else:
            return result


class PublicBoolean(dsl.PublicBoolean):

    def __init__(self, value):
        super().__init__(value.inner)

    def if_else(
        self: dsl.SecretBoolean,
        arg_0: _NadaType | "SecretRational" | "Rational",
        arg_1: _NadaType | "SecretRational" | "Rational",
    ) -> Union[SecretInteger, SecretUnsignedInteger]:
        first_arg = arg_0
        second_arg = arg_1
        if isinstance(arg_0, (SecretRational, Rational)) and isinstance(
            arg_1, (SecretRational, Rational)
        ):
            # Both are SecretRational or Rational objects
            if arg_0.log_scale != arg_1.log_scale:
                raise ValueError("Cannot output values with different scales.")
            first_arg = arg_0.value
            second_arg = arg_1.value
        elif isinstance(arg_0, (Rational, SecretRational)) or isinstance(
            arg_1, (Rational, SecretRational)
        ):
            # Both are SecretRational or Rational objects but of different type
            raise TypeError(f"Invalid operation: {self}.IfElse({arg_0}, {arg_1})")

        result = super().if_else(first_arg, second_arg)

        if isinstance(arg_0, (SecretRational, Rational)):
            # If we have a SecretBoolean, the return type will be SecretInteger, thus promoted to SecretRational
            return Rational.from_parts(result, arg_0.log_scale)
        else:
            return result


class RationalConfig(object):

    LOG_SCALE = 16

    @classmethod
    def set_rational_scale(cls, scale: int) -> None:
        """Sets the global scaling factor.

        Args:
            scale (int): New scaling factor in number of bits.
        """
        cls.LOG_SCALE = scale


class Rational:
    """Wrapper class to store scaled Integer values representing a fixed-point number."""

    def __init__(
        self,
        value: _Number,
        log_scale=RationalConfig.LOG_SCALE,
        is_scaled: bool = False,
    ) -> None:
        """Initializes wrapper around Integer object.

        Args:
            value (_Number): The value to be represented as a Rational.
            log_scale (int): Quantization scaling factor. Defaults to RationalConfig.LOG_SCALE.
            is_scaled (bool, optional): Flag that represents whether the value is already scaled.
                Defaults to False.

        Raises:
            TypeError: Raised when a value of an incompatible type is passed.
        """
        if not isinstance(value, (_Number, _NadaInteger)):
            raise TypeError("Cannot instantiate Rational from type `%s`." % type(value))

        if not is_scaled:
            value *= 1 << log_scale
            value = round(value)

        self._log_scale = log_scale
        if isinstance(value, _NadaInteger):
            self._value = value
        else:
            self._value = Integer(value)

    @property
    def log_scale(self) -> int:
        """Getter for the logarithmic scale value.

        Returns:
            int: Logarithmic scale value.
        """
        return self._log_scale

    @property
    def value(self) -> Integer:
        """Getter for the underlying Integer value.

        Returns:
            Integer: The Integer value.
        """
        return self._value

    @classmethod
    def from_parts(cls, value: Integer, scale: int) -> "Rational":
        """Creates a Rational from an Integer value and scale.

        Args:
            value (Integer): Integer value to convert.
            scale (int): Quantization scaling factor.

        Returns:
            Rational: Instantiated wrapper around number.
        """
        return Rational(value, scale, is_scaled=True)

    def add(
        self, other: _NadaRational, unchecked: bool = False
    ) -> Union["Rational", "SecretRational"]:
        """Add two rational numbers.

        Args:
            other (_NadaRational): Other rational number to add.
            unchecked (bool, optional): Flag to disable scale checking. Defaults to False.

        Returns:
            Union[Rational, SecretRational]: Result of the addition.

        Raises:
            TypeError: If the other value is of an incompatible type.
        """
        if not isinstance(other, (Rational, SecretRational)):
            raise TypeError(
                f"Operation + not allowed between types {type(self)} + {type(other)}"
            )

        if not unchecked and self.log_scale != other.log_scale:
            raise ValueError("Cannot add values with different scales.")

        if isinstance(other, SecretRational):
            return SecretRational.from_parts(other.value + self.value, self.log_scale)
        else:
            return Rational.from_parts(self.value + other.value, self.log_scale)

    def __add__(self, other: _NadaRational) -> Union["Rational", "SecretRational"]:
        return self.add(other)

    def __iadd__(self, other: _NadaRational) -> Union["Rational", "SecretRational"]:
        return self.add(other)

    def sub(
        self, other: _NadaRational, unchecked: bool = False
    ) -> Union["Rational", "SecretRational"]:
        """Subtract two rational numbers.

        Args:
            other (_NadaRational): Other rational number to subtract.

        Returns:
            Union[Rational, SecretRational]: Result of the subtraction.
            unchecked (bool, optional): Flag to disable scale checking. Defaults to False.

        Raises:
            TypeError: If the other value is of an incompatible type.
        """
        if not isinstance(other, (Rational, SecretRational)):
            raise TypeError(
                f"Operation - not allowed between types {type(self)} - {type(other)}"
            )

        if not unchecked and self.log_scale != other.log_scale:
            raise ValueError("Cannot substract values with different scales.")

        if isinstance(other, SecretRational):
            return SecretRational.from_parts(self.value - other.value, self.log_scale)
        else:
            return Rational.from_parts(self.value - other.value, self.log_scale)

    def __sub__(self, other: _NadaRational) -> Union["Rational", "SecretRational"]:
        return self.sub(other)

    def __isub__(self, other: _NadaLike) -> Union["Rational", "SecretRational"]:
        return self.sub(other)

    def mul_no_rescale(
        self, other: _NadaRational, unchecked: bool = False
    ) -> Union["Rational", "SecretRational"]:
        """Multiply two rational numbers.

        WARNING: This function does not rescale by default. Use `mul` to multiply and rescale.

        Args:
            other (_NadaRational): Other rational number to multiply.
            unchecked (bool, optional): Flag to disable scale checking. Defaults to False.

        Returns:
            Union[Rational, SecretRational]: Result of the multiplication.

        Raises:
            TypeError: If the other value is of an incompatible type.
        """
        if not isinstance(other, (Rational, SecretRational)):
            raise TypeError(
                f"Operation * not allowed between types {type(self)} * {type(other)}"
            )

        if not unchecked and self.log_scale != other.log_scale:
            raise ValueError("Cannot multiply values with different scales.")

        if isinstance(other, SecretRational):
            return SecretRational.from_parts(
                self.value * other.value, self.log_scale + other.log_scale
            )
        else:
            return Rational.from_parts(
                self.value * other.value, self.log_scale + other.log_scale
            )

    def mul(
        self, other: _NadaRational, unchecked: bool = False
    ) -> Union["Rational", "SecretRational"]:
        """Multiply two rational numbers and rescale the result.

        Args:
            other (_NadaRational): Other rational number to multiply.
            unchecked (bool, optional): Flag to disable scale checking. Defaults to False.

        Returns:
            Union[Rational, SecretRational]: Result of the multiplication.
        """
        c = self.mul_no_rescale(other, unchecked=unchecked)
        d = c.rescale_down()
        return d

    def __mul__(self, other: _NadaLike) -> Union["Rational", "SecretRational"]:
        return self.mul(other)

    def __imul__(self, other: _NadaLike) -> Union["Rational", "SecretRational"]:
        return self.mul(other)

    def divide_no_rescale(
        self, other: _NadaRational, unchecked: bool = False
    ) -> Union["Rational", "SecretRational"]:
        """Divide two rational numbers.

        Args:
            other (_NadaRational): Other rational number to divide by.

        Returns:
            Union[Rational, SecretRational]: Result of the division.

        Raises:
            TypeError: If the other value is of an incompatible type.
        """
        if not isinstance(other, (Rational, SecretRational)):
            raise TypeError(
                f"Operation / not allowed between types {type(self)} / {type(other)}"
            )

        if (
            not unchecked
            and self.log_scale != other.log_scale + RationalConfig.LOG_SCALE
        ):
            raise ValueError(
                f"Cannot divide values where scale is: {self.log_scale} / {other.log_scale}. Required scale: {self.log_scale}  / {other.log_scale + RationalConfig.LOG_SCALE}"
            )

        if isinstance(other, SecretRational):
            return SecretRational.from_parts(
                self.value / other.value, self.log_scale - other.log_scale
            )
        else:
            return Rational.from_parts(
                self.value / other.value, self.log_scale - other.log_scale
            )

    def divide(self, other: _NadaRational) -> Union["Rational", "SecretRational"]:
        """Divide two rational numbers and rescale the result.

        Args:
            other (_NadaRational): Other rational number to divide by.

        Returns:
            Union[Rational, SecretRational]: Result of the division.
        """
        a = self.rescale_up()
        c = a.divide_no_rescale(other)
        return c

    def __truediv__(self, other: _NadaRational) -> Union["Rational", "SecretRational"]:
        return self.divide(other)

    def __rtruediv__(self, other: _NadaRational) -> Union["Rational", "SecretRational"]:
        return self.divide(other)

    def __itruediv__(self, other: _NadaRational) -> Union["Rational", "SecretRational"]:
        return self.divide(other)

    def __pow__(self, other: int) -> "Rational":
        """Raise a rational number to an integer power using binary exponentiation.

        Args:
            other (int): The exponent.

        Returns:
            Rational: Result of the power operation.

        Raises:
            TypeError: If the exponent is not an integer.
        """
        if not isinstance(other, int):
            raise TypeError(
                "Cannot raise Rational to a power of type `%s`" % type(other)
            )

        result = Rational(1, self.log_scale)

        if other == 0:
            return result  # Any number to the power of 0 is 1

        base = self

        exponent = abs(other)
        while exponent > 0:
            if exponent % 2 == 1:
                result = result * base
            base *= base
            exponent //= 2

        if other < 0:
            return Rational.from_parts(1 / result.value, result.log_scale)

        return result

    def __neg__(self) -> "SecretRational":
        """Negate the SecretRational value.

        Returns:
            SecretRational: Negated SecretRational value.
        """
        return Rational.from_parts(self.value * Integer(-1), self.log_scale)

    def __lt__(self, other: _NadaRational) -> SecretBoolean:
        """Check if this SecretRational is less than another.

        Args:
            other (_NadaRational): The other SecretRational to compare against.

        Raises:
            ValueError: If the log scales of the two values are different.

        Returns:
            SecretBoolean: Result of the comparison.
        """
        if self.log_scale != other.log_scale:
            raise ValueError("Cannot compare values with different scales.")
        return self.value < other.value

    def __gt__(self, other: _NadaRational) -> SecretBoolean:
        """Check if this SecretRational is greater than another.

        Args:
            other (_NadaRational): The other SecretRational to compare against.

        Raises:
            ValueError: If the log scales of the two values are different.

        Returns:
            SecretBoolean: Result of the comparison.
        """
        if self.log_scale != other.log_scale:
            raise ValueError("Cannot compare values with different scales.")
        return self.value > other.value

    def __le__(self, other: _NadaRational) -> SecretBoolean:
        """Check if this SecretRational is less than or equal to another.

        Args:
            other (_NadaRational): The other SecretRational to compare against.

        Raises:
            ValueError: If the log scales of the two values are different.

        Returns:
            SecretBoolean: Result of the comparison.
        """
        if self.log_scale != other.log_scale:
            raise ValueError("Cannot compare values with different scales.")
        return self.value <= other.value

    def __ge__(self, other: _NadaRational) -> SecretBoolean:
        """Check if this SecretRational is greater than or equal to another.

        Args:
            other (_NadaRational): The other SecretRational to compare against.

        Raises:
            ValueError: If the log scales of the two values are different.

        Returns:
            SecretBoolean: Result of the comparison.
        """
        if self.log_scale != other.log_scale:
            raise ValueError("Cannot compare values with different scales.")
        return self.value >= other.value

    def __eq__(self, other: _NadaRational) -> SecretBoolean:
        """Check if this SecretRational is equal to another.

        Args:
            other (_NadaRational): The other SecretRational to compare against.

        Raises:
            ValueError: If the log scales of the two values are different.

        Returns:
            SecretBoolean: Result of the comparison.
        """
        if self.log_scale != other.log_scale:
            raise ValueError("Cannot compare values with different scales.")
        return self.value == other.value

    def rescale_up(self, log_scale: int = RationalConfig.LOG_SCALE):
        """Rescale the value in the upward direction by a scaling factor (default RationalConfig.LOG_SCALE).

        This is equivalent to multiplying the value by `2**(log_scale)`.

        Args:
            log_scale (int): Scaling factor to rescale the value.

        Returns:
            SecretRational: Rescaled SecretRational value.
        """
        try:
            return Rational.from_parts(
                self._value << UnsignedInteger(log_scale), self.log_scale + log_scale
            )
        except TypeError:
            return Rational.from_parts(
                self._value * Integer(1 << log_scale), self.log_scale + log_scale
            )

    def rescale_down(self, log_scale: int = RationalConfig.LOG_SCALE):
        """Rescale the value in the downward direction by a scaling factor (default RationalConfig.LOG_SCALE).

        This is equivalent to dividing the value by `2**(log_scale)`.

        Args:
            log_scale (int): Scaling factor to rescale the value.

        Returns:
            SecretRational: Rescaled SecretRational value.
        """
        try:
            return Rational.from_parts(
                self._value >> UnsignedInteger(log_scale), self.log_scale - log_scale
            )
        except TypeError:
            return Rational.from_parts(
                self._value / Integer(1 << log_scale), self.log_scale - log_scale
            )


class SecretRational:
    """Wrapper class to store scaled SecretInteger values representing a fixed-point number."""

    def __init__(
        self,
        name: str = None,
        party: Party = None,
        value: SecretInteger = None,
        log_scale=RationalConfig.LOG_SCALE,
    ) -> None:
        """Initializes wrapper around SecretInteger object. The object should come scaled up by default otherwise precision may be lost.

        Args:
            name (str, optional): Name for the SecretInteger input.
            party (Party, optional): Party associated with the SecretInteger input.
            value (SecretInteger, optional): SecretInteger value to be represented as SecretRational.
            log_scale (int): Quantization scaling factor. Defaults to RationalConfig.LOG_SCALE.

        Raises:
            ValueError: If neither value nor name and party are provided.
        """
        if value is not None:
            self._value = value
            self._log_scale = log_scale

        elif name is not None and party is not None:
            self._value = SecretInteger(Input(name=name, party=party))
            self._log_scale = log_scale

        else:
            raise ValueError(
                "Either value or name and party must be provided or value must be a SecretInteger."
            )

    @property
    def log_scale(self) -> int:
        """Getter for the logarithmic scale value.

        Returns:
            int: Logarithmic scale value.
        """
        return self._log_scale

    @property
    def value(self) -> SecretInteger:
        """Getter for the underlying SecretInteger value.

        Returns:
            SecretInteger: The SecretInteger value.
        """
        return self._value

    @classmethod
    def from_parts(cls, value: Integer, log_scale: int) -> "SecretRational":
        """Creates a SecretRational from an Integer value.

        Args:
            value (Integer): Integer value to convert.
            log_scale (int): Quantization scaling factor.

        Returns:
            SecretRational: Instantiated SecretRational object.
        """
        return SecretRational(value=value, log_scale=log_scale)

    def add(self, other: _NadaRational, unchecked: bool = False) -> "SecretRational":
        """Add two SecretRational numbers.

        Args:
            other (_NadaRational): The other SecretRational to add.
            unchecked (bool, optional): Flag to disable scale checking. Defaults to False.

        Raises:
            TypeError: If the other value is not a Rational or SecretRational.

        Returns:
            SecretRational: Result of the addition.
        """
        if not isinstance(other, (Rational, SecretRational)):
            raise TypeError(
                f"Operation + not allowed between types {type(self)} + {type(other)}"
            )

        if not unchecked and self.log_scale != other.log_scale:
            raise ValueError("Cannot add values with different scales.")

        return SecretRational.from_parts(self.value + other.value, self.log_scale)

    def __add__(self, other: _NadaRational) -> "SecretRational":
        """Override the addition operator for SecretRational.

        Args:
            other (_NadaRational): The other SecretRational to add.

        Returns:
            SecretRational: Result of the addition.
        """
        return self.add(other)

    def __iadd__(self, other: _NadaRational) -> "SecretRational":
        """Override the in-place addition operator for SecretRational.

        Args:
            other (_NadaRational): The other SecretRational to add.

        Returns:
            SecretRational: Result of the addition.
        """
        return self.add(other)

    def sub(self, other: _NadaRational, unchecked: bool = False) -> "SecretRational":
        """Subtract two SecretRational numbers.

        Args:
            other (_NadaRational): The other SecretRational to subtract.
            unchecked (bool, optional): Flag to disable scale checking. Defaults to False.

        Raises:
            TypeError: If the other value is not a Rational or SecretRational.

        Returns:
            SecretRational: Result of the subtraction.
        """
        if not isinstance(other, (Rational, SecretRational)):
            raise TypeError(
                f"Operation - not allowed between types {type(self)} - {type(other)}"
            )
        if not unchecked and self.log_scale != other.log_scale:
            raise ValueError("Cannot substract values with different scales.")

        return SecretRational.from_parts(self.value - other.value, self.log_scale)

    def __sub__(self, other: _NadaRational) -> "SecretRational":
        """Override the subtraction operator for SecretRational.

        Args:
            other (_NadaRational): The other SecretRational to subtract.

        Returns:
            SecretRational: Result of the subtraction.
        """
        return self.sub(other)

    def __isub__(self, other: _NadaLike) -> "SecretRational":
        """Override the in-place subtraction operator for SecretRational.

        Args:
            other (_NadaLike): The other value to subtract.

        Returns:
            SecretRational: Result of the subtraction.
        """
        return self.sub(other)

    def mul_no_rescale(
        self, other: _NadaRational, unchecked: bool = False
    ) -> "SecretRational":
        """Multiply two SecretRational numbers.

        Args:
            other (_NadaRational): The other SecretRational to multiply.
            unchecked (bool, optional): Flag to disable scale checking. Defaults to False.


        Raises:
            TypeError: If the other value is not a Rational or SecretRational.

        Returns:
            SecretRational: Result of the multiplication.
        """
        if not isinstance(other, (Rational, SecretRational)):
            raise TypeError(
                f"Operation * not allowed between types {type(self)} * {type(other)}"
            )

        if not unchecked and self.log_scale != other.log_scale:
            raise ValueError("Cannot multiply values with different scales.")

        return SecretRational.from_parts(
            self.value * other.value, self.log_scale + other.log_scale
        )

    def mul(self, other: _NadaRational, unchecked: bool = False) -> "SecretRational":
        """Multiply two SecretRational numbers and rescale the result.

        Args:
            other (_NadaRational): The other SecretRational to multiply.
            unchecked (bool, optional): Flag to disable scale checking. Defaults to False.

        Returns:
            SecretRational: Result of the multiplication, rescaled.
        """
        c = self.mul_no_rescale(other, unchecked=unchecked)
        d = c.rescale_down()
        return d

    def __mul__(self, other: _NadaLike) -> "SecretRational":
        """Override the multiplication operator for SecretRational.

        Args:
            other (_NadaLike): The other value to multiply.

        Returns:
            SecretRational: Result of the multiplication, rescaled.
        """
        return self.mul(other)

    def __imul__(self, other: _NadaLike) -> "SecretRational":
        """Override the in-place multiplication operator for SecretRational.

        Args:
            other (_NadaLike): The other value to multiply.

        Returns:
            SecretRational: Result of the multiplication, rescaled.
        """
        return self.mul(other)

    def divide_no_rescale(
        self, other: _NadaRational, unchecked: bool = False
    ) -> "SecretRational":
        """Divide two SecretRational numbers.

        Args:
            other (_NadaRational): The other SecretRational to divide by.
            unchecked (bool, optional): Flag to disable scale checking. Defaults to False.

        Raises:
            TypeError: If the other value is not a Rational or SecretRational.

        Returns:
            SecretRational: Result of the division.
        """
        if not isinstance(other, (Rational, SecretRational)):
            raise TypeError(
                f"Operation / not allowed between types {type(self)} / {type(other)}"
            )

        if (
            not unchecked
            and self.log_scale != other.log_scale + RationalConfig.LOG_SCALE
        ):
            raise ValueError(
                f"Cannot divide values where scale is: {self.log_scale} / {other.log_scale}. Required scale: {self.log_scale}  / {other.log_scale + RationalConfig.LOG_SCALE}"
            )

        return SecretRational.from_parts(
            self.value / other.value, self.log_scale - other.log_scale
        )

    def divide(self, other: _NadaRational, unchecked: bool = False) -> "SecretRational":
        """Divide two SecretRational numbers and rescale the result.

        Args:
            other (_NadaRational): The other SecretRational to divide by.
            unchecked (bool, optional): Flag to disable scale checking. Defaults to False.

        Returns:
            SecretRational: Result of the division, rescaled.
        """
        a = self.rescale_up()
        c = a.divide_no_rescale(other, unchecked=unchecked)
        return c

    def __truediv__(self, other: _NadaRational) -> "SecretRational":
        """Override the true division operator for SecretRational.

        Args:
            other (_NadaRational): The other value to divide by.

        Returns:
            SecretRational: Result of the division, rescaled.
        """
        return self.divide(other)

    def __rtruediv__(self, other: _NadaRational) -> "SecretRational":
        """Override the reflected true division operator for SecretRational.

        Args:
            other (_NadaRational): The other value to divide by.

        Returns:
            SecretRational: Result of the division, rescaled.
        """
        return self.divide(other)

    def __itruediv__(self, other: _NadaRational) -> "SecretRational":
        """Override the in-place true division operator for SecretRational.

        Args:
            other (_NadaRational): The other value to divide by.

        Returns:
            SecretRational: Result of the division, rescaled.
        """
        return self.divide(other)

    def __pow__(self, other: int) -> "SecretRational":
        """Raise a SecretRational to an integer power using binary exponentiation.

        Args:
            other (int): The exponent.

        Raises:
            TypeError: If the exponent is not an integer.

        Returns:
            SecretRational: Result of the power operation.
        """
        if not isinstance(other, int):
            raise TypeError(
                "Cannot raise SecretRational to a power of type `%s`" % type(other)
            )

        result = Rational(1, self.log_scale)

        if other == 0:
            return result  # Any number to the power of 0 is 1

        base = self

        exponent = abs(other)
        while exponent > 0:
            if exponent % 2 == 1:
                result = result * base
            base *= base
            exponent //= 2

        if other < 0:
            return SecretRational.from_parts(1 / result.value, result.log_scale)

        return result

    def __neg__(self) -> "SecretRational":
        """Negate the SecretRational value.

        Returns:
            SecretRational: Negated SecretRational value.
        """
        return SecretRational.from_parts(self.value * Integer(-1), self.log_scale)

    def __lshift__(self, other: _NadaLike) -> "SecretRational":
        """Left shift the SecretRational value.

        Args:
            other (_NadaLike): The value to left shift by.

        Returns:
            SecretRational: Left shifted SecretRational value.
        """
        return SecretRational(self.value << other, self.log_scale)

    def __rshift__(self, other: _NadaLike) -> "SecretRational":
        """Right shift the SecretRational value.

        Args:
            other (_NadaLike): The value to right shift by.

        Returns:
            SecretRational: Right shifted SecretRational value.
        """
        return SecretRational(self.value >> other, self.log_scale)

    def __lt__(self, other: _NadaRational) -> SecretBoolean:
        """Check if this SecretRational is less than another.

        Args:
            other (_NadaRational): The other SecretRational to compare against.

        Raises:
            ValueError: If the log scales of the two values are different.

        Returns:
            SecretBoolean: Result of the comparison.
        """
        if self.log_scale != other.log_scale:
            raise ValueError("Cannot compare values with different scales.")
        return SecretBoolean(self.value < other.value)

    def __gt__(self, other: _NadaRational) -> SecretBoolean:
        """Check if this SecretRational is greater than another.

        Args:
            other (_NadaRational): The other SecretRational to compare against.

        Raises:
            ValueError: If the log scales of the two values are different.

        Returns:
            SecretBoolean: Result of the comparison.
        """
        if self.log_scale != other.log_scale:
            raise ValueError("Cannot compare values with different scales.")
        return SecretBoolean(self.value > other.value)

    def __le__(self, other: _NadaRational) -> SecretBoolean:
        """Check if this SecretRational is less than or equal to another.

        Args:
            other (_NadaRational): The other SecretRational to compare against.

        Raises:
            ValueError: If the log scales of the two values are different.

        Returns:
            SecretBoolean: Result of the comparison.
        """
        if self.log_scale != other.log_scale:
            raise ValueError("Cannot compare values with different scales.")
        return SecretBoolean(self.value <= other.value)

    def __ge__(self, other: _NadaRational) -> SecretBoolean:
        """Check if this SecretRational is greater than or equal to another.

        Args:
            other (_NadaRational): The other SecretRational to compare against.

        Raises:
            ValueError: If the log scales of the two values are different.

        Returns:
            SecretBoolean: Result of the comparison.
        """
        if self.log_scale != other.log_scale:
            raise ValueError("Cannot compare values with different scales.")
        return SecretBoolean(self.value >= other.value)

    def __eq__(self, other: _NadaRational) -> SecretBoolean:
        """Check if this SecretRational is equal to another.

        Args:
            other (_NadaRational): The other SecretRational to compare against.

        Raises:
            ValueError: If the log scales of the two values are different.

        Returns:
            SecretBoolean: Result of the comparison.
        """
        if self.log_scale != other.log_scale:
            raise ValueError("Cannot compare values with different scales.")
        return SecretBoolean(self.value == other.value)

    def __ne__(self, other: _NadaRational) -> SecretBoolean:
        """Check if this SecretRational is not equal to another.

        Args:
            other (_NadaRational): The other SecretRational to compare against.

        Raises:
            ValueError: If the log scales of the two values are different.

        Returns:
            SecretBoolean: Result of the comparison.
        """
        if self.log_scale != other.log_scale:
            raise ValueError("Cannot compare values with different scales.")
        return not (self.value == other.value)

    def public_equals(self, other: _NadaRational) -> PublicBoolean:
        """Check if this SecretRational is equal to another and reveal the result.

        Args:
            other (_NadaRational): The other SecretRational to compare against.

        Raises:
            ValueError: If the log scales of the two values are different.

        Returns:
            PublicBoolean: Result of the comparison, revealed.
        """
        if self.log_scale != other.log_scale:
            raise ValueError("Cannot compare values with different scales.")
        return Rational.from_parts(
            self.value.public_equals(other.value), self.log_scale
        )

    def reveal(self) -> Rational:
        """Reveal the SecretRational value.

        Returns:
            Rational: Revealed SecretRational value.
        """
        return SecretRational.from_parts(self.value.reveal(), self.log_scale)

    def trunc_pr(self, arg_0: _NadaLike) -> "SecretRational":
        """Truncate the SecretRational value.

        Args:
            arg_0 (_NadaLike): The value to truncate by.

        Returns:
            SecretRational: Truncated SecretRational value.
        """
        return SecretRational.from_parts(self.value.trunc_pr(arg_0), self.log_scale)

    def rescale_up(self, log_scale: int = RationalConfig.LOG_SCALE) -> "SecretRational":
        """Rescale the SecretRational value upwards by a scaling factor.

        Args:
            log_scale (int): The scaling factor.

        Returns:
            SecretRational: Rescaled SecretRational value.
        """
        try:
            return SecretRational.from_parts(
                self._value << UnsignedInteger(log_scale), self.log_scale + log_scale
            )
        except TypeError:
            return SecretRational.from_parts(
                self._value * Integer(1 << log_scale), self.log_scale + log_scale
            )

    def rescale_down(
        self, log_scale: int = RationalConfig.LOG_SCALE
    ) -> "SecretRational":
        """Rescale the SecretRational value downwards by a scaling factor.

        Args:
            log_scale (int): The scaling factor.

        Returns:
            SecretRational: Rescaled SecretRational value.
        """
        try:
            return SecretRational.from_parts(
                self._value >> UnsignedInteger(log_scale), self.log_scale - log_scale
            )
        except TypeError:
            return SecretRational.from_parts(
                self._value / Integer(1 << log_scale), self.log_scale - log_scale
            )



"""
    This module provides an enhanced array class, NadaArray, with additional functionality
    for mathematical operations and integration with the Nada Algebra library.
"""

from dataclasses import dataclass
from typing import Any, Callable, Union

import numpy as np
from nada_dsl import (
    Input,
    Party,
    Output,
    SecretInteger,
    SecretUnsignedInteger,
    PublicInteger,
    PublicUnsignedInteger,
    Integer,
    UnsignedInteger,
)
from nada_algebra.types import Rational, SecretRational, RationalConfig

_NadaOperand = Union[
    "NadaArray",
    np.ndarray,
    int,
    Integer,
    UnsignedInteger,
    SecretInteger,
    SecretUnsignedInteger,
    PublicInteger,
    PublicUnsignedInteger,
    Rational,
    SecretRational,
]


@dataclass
class NadaArray:
    """
    Represents an array-like object with additional functionality.

    Attributes:
        inner (np.ndarray): The underlying NumPy array.
    """

    inner: np.ndarray

    SUPPORTED_OPERATIONS = {
        "base",
        "compress",
        "copy",
        "cumprod",
        "cumsum",
        "data",
        "diagonal",
        "dtype",
        "fill",
        "flags",
        "flat",
        "flatten",
        "item",
        "itemset",
        "itemsize",
        "nbytes",
        "ndim",
        "prod",
        "put",
        "ravel",
        "repeat",
        "reshape",
        "resize",
        "shape",
        "size",
        "squeeze",
        "strides",
        "sum",
        "swapaxes",
        "T",
        "take",
        "tolist",
        "trace",
        "transpose",
    }

    def __getitem__(self, item):
        """
        Get an item from the array.

        Args:
            item: The item to retrieve.

        Returns:
            NadaArray: A new NadaArray representing the retrieved item.
        """
        if len(self.inner.shape) == 1:
            return self.inner[item]
        return NadaArray(self.inner[item])

    def __setitem__(self, key, value):
        """
        Set an item in the array.

        Args:
            key: The key to set.
            value: The value to set.
        """
        if isinstance(value, NadaArray):
            self.inner[key] = value.inner
        else:
            self.inner[key] = value

    def __add__(self, other: _NadaOperand) -> "NadaArray":
        """
        Perform element-wise addition with broadcasting.

        Args:
            other (_NadaOperand): The object to add.

        Returns:
            NadaArray: A new NadaArray representing the element-wise addition result.
        """
        if isinstance(other, NadaArray):
            return NadaArray(self.inner + other.inner)
        if isinstance(other, int):
            return NadaArray(self.inner + Integer(other))
        return NadaArray(self.inner + other)

    def __sub__(self, other: _NadaOperand) -> "NadaArray":
        """
        Perform element-wise subtraction with broadcasting.

        Args:
            other (_NadaOperand): The object to subtract.

        Returns:
            NadaArray: A new NadaArray representing the element-wise subtraction result.
        """
        if isinstance(other, NadaArray):
            return NadaArray(self.inner - other.inner)
        if isinstance(other, int):
            return NadaArray(self.inner - Integer(other))
        return NadaArray(self.inner - other)

    def __mul__(self, other: _NadaOperand) -> "NadaArray":
        """
        Perform element-wise multiplication with broadcasting.

        Args:
            other (_NadaOperand): The object to multiply.

        Returns:
            NadaArray: A new NadaArray representing the element-wise multiplication result.
        """
        if isinstance(other, NadaArray):
            return NadaArray(self.inner * other.inner)
        if isinstance(other, int):
            return NadaArray(self.inner * Integer(other))
        return NadaArray(self.inner * other)

    def __pow__(self, other: int) -> "NadaArray":
        """Raises NadaArray to a power.

        Args:
            other (int): Power value.

        Returns:
            NadaArray: Result NadaArray.
        """
        if not isinstance(other, int):
            raise TypeError(
                "Cannot raise `NadaArray` to power of type `%s`" % type(other).__name__
            )
        result = self.copy()
        for _ in range(other - 1):
            result = result * result
        return result

    def __truediv__(self, other: _NadaOperand) -> "NadaArray":
        """
        Perform element-wise division with broadcasting.

        Args:
            other (_NadaOperand): The object to divide.

        Returns:
            NadaArray: A new NadaArray representing the element-wise division result.
        """
        if isinstance(other, NadaArray):
            return NadaArray(self.inner / other.inner)
        if isinstance(other, int):
            return NadaArray(self.inner / Integer(other))
        return NadaArray(self.inner / other)

    def __matmul__(self, other: "NadaArray") -> "NadaArray":
        """
        Perform matrix multiplication with another NadaArray.

        Args:
            other (NadaArray): The NadaArray to perform matrix multiplication with.

        Returns:
            NadaArray: A new NadaArray representing the result of matrix multiplication.
        """
        result = self.inner @ other.inner
        return NadaArray(np.array(result))

    @property
    def ndim(self) -> int:
        return len(self.shape)

    def dot(self, other: "NadaArray") -> "NadaArray":
        """
        Compute the dot product between two NadaArray objects.

        Args:
            other (NadaArray): The NadaArray to compute dot product with.

        Returns:
            NadaArray: A new NadaArray representing the dot product result.
        """
        return NadaArray(self.inner.dot(other.inner))

    def hstack(self, other: "NadaArray") -> "NadaArray":
        """
        Horizontally stack two NadaArray objects.

        Args:
            other (NadaArray): The NadaArray to stack horizontally with.

        Returns:
            NadaArray: A new NadaArray representing the horizontal stack.
        """
        return NadaArray(np.hstack((self.inner, other.inner)))

    def vstack(self, other: "NadaArray") -> "NadaArray":
        """
        Vertically stack two NadaArray objects.

        Args:
            other (NadaArray): The NadaArray to stack vertically with.

        Returns:
            NadaArray: A new NadaArray representing the vertical stack.
        """
        return NadaArray(np.vstack((self.inner, other.inner)))

    def reveal(self) -> "NadaArray":
        """
        Reveal the elements of the array.

        Returns:
            NadaArray: A new NadaArray with revealed values.
        """
        return self.applypyfunc(lambda x: x.reveal())

    @staticmethod
    def apply_function_elementwise(func: Callable, array: np.ndarray) -> np.ndarray:
        """
        Apply a function element-wise to the input array.

        Args:
            func (Callable): The function to apply.
            array (np.ndarray): The input array.

        Returns:
            np.ndarray: A NumPy array with the function applied to each element.
        """
        if len(array.shape) == 1:
            return [func(x) for x in array]
        return [
            NadaArray.apply_function_elementwise(func, array[i])
            for i in range(array.shape[0])
        ]

    def applypyfunc(self, func: Callable) -> "NadaArray":
        """
        Apply a Python function element-wise to the array.

        Args:
            func (Callable): The function to apply.

        Returns:
            NadaArray: A new NadaArray with the function applied to each element.
        """
        return NadaArray(
            np.array(NadaArray.apply_function_elementwise(func, self.inner))
        )

    @staticmethod
    def output_array(array: np.ndarray, party: Party, prefix: str) -> list:
        """
        Generate a list of Output objects for each element in the input array.

        Args:
            array (np.ndarray): The input array.
            party (Party): The party object.
            prefix (str): The prefix to be added to the Output names.

        Returns:
            list: A list of Output objects.
        """
        if isinstance(
            array,
            (
                SecretInteger,
                SecretUnsignedInteger,
                PublicInteger,
                PublicUnsignedInteger,
                Integer,
                UnsignedInteger,
            ),
        ):
            return [Output(array, f"{prefix}_0", party)]
        elif isinstance(array, (Rational, SecretRational)):
            return [Output(array.value, f"{prefix}_0", party)]

        if len(array.shape) == 1:
            return [
                (
                    Output(array[i].value, f"{prefix}_{i}", party)
                    if isinstance(array[i], (Rational, SecretRational))
                    else Output(array[i], f"{prefix}_{i}", party)
                )
                for i in range(array.shape[0])
            ]
        return [
            v
            for i in range(array.shape[0])
            for v in NadaArray.output_array(array[i], party, f"{prefix}_{i}")
        ]

    def output(self, party: Party, prefix: str) -> list:
        """
        Generate a list of Output objects for each element in the NadaArray.

        Args:
            party (Party): The party object.
            prefix (str): The prefix for naming the Output objects.

        Returns:
            list: A list of Output objects.
        """
        return NadaArray.output_array(self.inner, party, prefix)

    @staticmethod
    def create_list(dims: list, party: Party, prefix: str, generator: Callable) -> list:
        """
        Recursively create a nested list of objects generated by a generator function.

        Args:
            dims (list): A list of integers representing the dimensions of the desired nested list.
            party (Party): The party object representing the party to which the elements belong.
            prefix (str): A prefix for the names of the elements.
            generator (Callable): The function to generate the elements.

        Returns:
            list: A nested list of generated objects.
        """
        if len(dims) == 1:
            return [generator(f"{prefix}_{i}", party) for i in range(dims[0])]
        return [
            NadaArray.create_list(
                dims[1:],
                party,
                f"{prefix}_{i}",
                generator,
            )
            for i in range(dims[0])
        ]

    @staticmethod
    def array(
        dims: list,
        party: Party,
        prefix: str,
        nada_type: Union[
            SecretInteger,
            SecretUnsignedInteger,
            PublicInteger,
            PublicUnsignedInteger,
            SecretRational,
            Rational,
        ] = SecretInteger,
    ) -> "NadaArray":
        """
        Create a NadaArray with the specified dimensions and element type.

        Args:
            dims (list): A list of integers representing the dimensions of the array.
            party (Party): The party object.
            prefix (str): A prefix for naming the array elements.
            nada_type (type, optional): The type of elements to create. Defaults to SecretInteger.

        Returns:
            NadaArray: The created NadaArray.

        Raises:
            ValueError: Raised if the nada_type is not supported.
        """
        generator = None
        if nada_type in (Rational, SecretRational):
            generator = lambda name, party: nada_type(name=name, party=party)
        elif nada_type in (
            SecretInteger,
            SecretUnsignedInteger,
            PublicInteger,
            PublicUnsignedInteger,
        ):
            generator = lambda name, party: nada_type(Input(name=name, party=party))
        else:
            raise ValueError(f"Unsupported nada_type: {nada_type}")

        return NadaArray(
            np.array(NadaArray.create_list(dims, party, prefix, generator))
        )

    @staticmethod
    def random(
        dims: list,
        nada_type: Union[
            SecretInteger, SecretUnsignedInteger, SecretRational
        ] = SecretInteger,
    ) -> "NadaArray":
        """
        Create a random NadaArray with the specified dimensions.

        Args:
            dims (list): A list of integers representing the dimensions of the array.
            nada_type (type, optional): The type of elements to create. Defaults to SecretInteger.

        Returns:
            NadaArray: The created random NadaArray.

        Raises:
            ValueError: Raised if the nada_type is not supported.
        """
        generator = None
        if nada_type is SecretRational:
            generator = lambda name, party: SecretRational.from_parts(
                SecretInteger.random(), RationalConfig.LOG_SCALE
            )
        elif nada_type in (SecretInteger, SecretUnsignedInteger):
            generator = lambda name, party: nada_type.random()
        else:
            raise ValueError(f"Unsupported nada_type: {nada_type}")

        return NadaArray(np.array(NadaArray.create_list(dims, None, None, generator)))

    def __getattr__(self, name: str) -> Any:
        """Routes other attributes to the inner NumPy array.

        Args:
            name (str): Attribute name.

        Raises:
            AttributeError: Raised if attribute not supported.

        Returns:
            Any: Result of attribute.
        """
        if name not in self.SUPPORTED_OPERATIONS:
            raise AttributeError(
                "NumPy method `%s` is not (currently) supported by NadaArrays." % name
            )

        attr = getattr(self.inner, name)

        if callable(attr):

            def wrapper(*args, **kwargs):
                result = attr(*args, **kwargs)
                if isinstance(result, np.ndarray):
                    return NadaArray(result)
                return result

            return wrapper

        if isinstance(attr, np.ndarray):
            attr = NadaArray(attr)

        return attr

    def __setattr__(self, name: str, value: Any):
        """
        Overrides the default behavior of setting attributes.

        If the attribute name is "inner", it sets the attribute value directly.
        Otherwise, it sets the attribute value on the inner object.

        Args:
            name (str): The name of the attribute.
            value: The value to set for the attribute.
        """
        if name == "inner":
            super().__setattr__(name, value)
        else:
            setattr(self.inner, name, value)

    def __len__(self):
        """
        Overrides the default behavior of returning the length of the object.

        Returns:
            int: The length of the inner numpy array.
        """
        return len(self.inner)



"""
This module provides functions to work with the Python Nillion Client for handling
secret and public variable integers and generating named party objects and input dictionaries.
"""

from typing import Union
from py_nillion_client import (
    SecretInteger,
    SecretUnsignedInteger,
    PublicVariableInteger,
    PublicVariableUnsignedInteger,
)
import numpy as np
from nada_algebra import types


def parties(num: int, prefix: str = "Party") -> list:
    """
    Creates a list of party name strings.

    Args:
        num (int): The number of parties to create.
        prefix (str, optional): The prefix to use for party names. Defaults to "Party".

    Returns:
        list: A list of party name strings in the format "{prefix}{i}".
    """
    return [f"{prefix}{i}" for i in range(num)]


def array(
    arr: np.ndarray,
    prefix: str,
    nada_type: Union[
        SecretInteger,
        SecretUnsignedInteger,
        PublicVariableInteger,
        PublicVariableUnsignedInteger,
        types.Rational,
        types.SecretRational,
    ] = SecretInteger,
) -> dict:
    """
    Recursively generates a dictionary of Nillion input objects for each element
    in the given array.

    Args:
        arr (np.ndarray): The input array.
        prefix (str): The prefix to be added to the output names.
        nada_type (type, optional): The type of the values introduced. Defaults to SecretInteger.

    Returns:
        dict: A dictionary mapping generated names to Nillion input objects.
    """
    # TODO: remove check for zero values when pushing zero secrets is supported
    if len(arr.shape) == 1:
        if nada_type == types.Rational:
            return {
                f"{prefix}_{i}": (PublicRational(arr[i])) for i in range(arr.shape[0])
            }
        if nada_type == types.SecretRational:
            return {
                f"{prefix}_{i}": (
                    SecretRational(arr[i]) if arr[i] != 0 else SecretInteger(1)
                )
                for i in range(arr.shape[0])
            }
        return {
            f"{prefix}_{i}": (
                nada_type(int(arr[i]))
                if (
                    nada_type in (PublicVariableInteger, PublicVariableUnsignedInteger)
                    or int(arr[i]) != 0
                )
                else nada_type(1)
            )
            for i in range(arr.shape[0])
        }
    return {
        k: v
        for i in range(arr.shape[0])
        for k, v in array(arr[i], f"{prefix}_{i}", nada_type).items()
    }


def concat(list_dict: list[dict]) -> dict:
    """
    Combines a list of dictionaries into a single dictionary.

    Note: This function will overwrite values for duplicate keys.

    Args:
        list_dict (list[dict]): A list of dictionaries.

    Returns:
        dict: A single merged dictionary.
    """
    return {k: v for d in list_dict for k, v in d.items()}


def __rational(value: Union[float, int]) -> int:
    """
    Returns the integer representation of the given float value.

    Args:
        value (Union[float, int]): The input value.

    Returns:
        int: The integer representation of the input value.
    """
    return round(value * (1 << types.RationalConfig.LOG_SCALE))


def PublicRational(value: Union[float, int]) -> PublicVariableInteger:
    """
    Returns the integer representation of the given float value.

    Args:
        value (Union[float, int]): The input value.

    Returns:
        int: The integer representation of the input value.
    """
    return PublicVariableInteger(__rational(value))


def SecretRational(value: Union[float, int]) -> SecretInteger:
    """
    Returns the integer representation of the given float value.

    Args:
        value (Union[float, int]): The input value.

    Returns:
        int: The integer representation of the input value.
    """
    return SecretInteger(__rational(value))



"""
This module provides common functions to work with Nada Algebra, including the creation
and manipulation of arrays and party objects.
"""

from typing import Any, Iterable, Union
from nada_dsl import (
    Party,
    SecretInteger,
    SecretUnsignedInteger,
    PublicInteger,
    PublicUnsignedInteger,
    Integer,
    UnsignedInteger,
)

import numpy as np
from nada_algebra.array import NadaArray
from nada_algebra.types import Rational, SecretRational


_NadaCleartextType = Union[Integer, UnsignedInteger, Rational]
""" A Nada Cleartext Type is: `Integer`, `UnsignedInteger`, or `Rational` """


def parties(num: int, prefix: str = "Party") -> list:
    """
    Create a list of Party objects with specified names.

    Args:
        num (int): The number of parties to create.
        prefix (str, optional): The prefix to use for party names. Defaults to "Party".

    Returns:
        list: A list of Party objects with names in the format "{prefix}{i}".
    """
    return [Party(name=f"{prefix}{i}") for i in range(num)]


def __from_numpy(arr: np.ndarray, nada_type: _NadaCleartextType) -> list:
    """
    Recursively convert a n-dimensional NumPy array to a nested list of NadaInteger objects.

    Args:
        arr (np.ndarray): A NumPy array of integers.
        nada_type (type): The type of NadaInteger objects to create.

    Returns:
        list: A nested list of NadaInteger objects.
    """
    if len(arr.shape) == 1:
        if isinstance(nada_type, Rational):
            return [nada_type(elem) for elem in arr]
        return [nada_type(int(elem)) for elem in arr]
    return [__from_numpy(arr[i], nada_type) for i in range(arr.shape[0])]


def from_list(lst: list, nada_type: _NadaCleartextType = Integer) -> NadaArray:
    """
    Create a cleartext NadaArray from a list of integers.

    Args:
        lst (list): A list of integers representing the elements of the array.
        nada_type (type, optional): The type of NadaInteger objects to create. Defaults to Integer.

    Returns:
        NadaArray: The created NadaArray.
    """
    if not isinstance(lst, np.ndarray):
        lst = np.array(lst)
    return NadaArray(np.array(__from_numpy(lst, nada_type)))


def ones(dims: Iterable[int], nada_type: _NadaCleartextType = Integer) -> NadaArray:
    """
    Create a cleartext NadaArray filled with ones.

    Args:
        dims (Iterable[int]): A list of integers representing the dimensions of the array.
        nada_type (type, optional): The type of NadaInteger objects to create. Defaults to Integer.

    Returns:
        NadaArray: The created NadaArray filled with ones.
    """
    return from_list(np.ones(dims), nada_type)


def ones_like(
    a: np.ndarray | NadaArray, nada_type: _NadaCleartextType = Integer
) -> NadaArray:
    """
    Create a cleartext NadaArray filled with one with the same shape and type as a given array.

    Args:
        a (np.ndarray | NadaArray): A reference array.
        nada_type (type, optional): The type of NadaInteger objects to create. Defaults to Integer.

    Returns:
        NadaArray: The created NadaArray filled with ones.
    """
    if isinstance(a, NadaArray):
        a = a.inner
    return from_list(np.ones_like(a), nada_type)


def zeros(dims: Iterable[int], nada_type: _NadaCleartextType = Integer) -> NadaArray:
    """
    Create a cleartext NadaArray filled with zeros.

    Args:
        dims (Iterable[int]): A list of integers representing the dimensions of the array.
        nada_type (type, optional): The type of NadaInteger objects to create. Defaults to Integer.

    Returns:
        NadaArray: The created NadaArray filled with zeros.
    """
    return from_list(np.zeros(dims), nada_type)


def zeros_like(
    a: np.ndarray | NadaArray, nada_type: _NadaCleartextType = Integer
) -> NadaArray:
    """
    Create a cleartext NadaArray filled with zeros with the same shape and type as a given array.

    Args:
        a (np.ndarray | NadaArray): A reference array.
        nada_type (type, optional): The type of NadaInteger objects to create. Defaults to Integer.

    Returns:
        NadaArray: The created NadaArray filled with zeros.
    """
    if isinstance(a, NadaArray):
        a = a.inner
    return from_list(np.zeros_like(a), nada_type)


def alphas(dims: Iterable[int], alpha: Any) -> NadaArray:
    """
    Create a NadaArray filled with a certain constant value.

    Args:
        dims (Iterable[int]): A list of integers representing the dimensions of the array.
        alpha (Any): Some constant value.

    Returns:
        NadaArray: NadaArray filled with constant value.
    """
    ones_array = np.ones(dims)
    return NadaArray(np.frompyfunc(lambda _: alpha, 1, 1)(ones_array))


def alphas_like(a: np.ndarray | NadaArray, alpha: Any) -> NadaArray:
    """
    Create a NadaArray filled with a certain constant value with the same shape and type as a given array.

    Args:
        a (np.ndarray | NadaArray): Reference array.
        alpha (Any): Some constant value.

    Returns:
        NadaArray: NadaArray filled with constant value.
    """
    if isinstance(a, NadaArray):
        a = a.inner
    ones_array = np.ones_like(a)
    return NadaArray(np.frompyfunc(lambda _: alpha, 1, 1)(ones_array))


def array(
    dims: Iterable[int],
    party: Party,
    prefix: str,
    nada_type: (
        SecretInteger
        | SecretUnsignedInteger
        | PublicInteger
        | PublicUnsignedInteger
        | SecretRational
    ) = SecretInteger,
) -> NadaArray:
    """
    Create a NadaArray with the specified dimensions and elements of the given type.

    Args:
        dims (Iterable[int]): A list of integers representing the dimensions of the array.
        party (Party): The party object.
        prefix (str): A prefix for naming the array elements.
        nada_type (type, optional): The type of elements to create. Defaults to SecretInteger.

    Returns:
        NadaArray: The created NadaArray.
    """
    return NadaArray.array(dims, party, prefix, nada_type)


def random(
    dims: Iterable[int],
    nada_type: SecretInteger | SecretUnsignedInteger | SecretRational = SecretInteger,
) -> NadaArray:
    """
    Create a random NadaArray with the specified dimensions.

    Args:
        dims (Iterable[int]): A list of integers representing the dimensions of the array.
        nada_type (type, optional): The type of elements to create. Defaults to SecretInteger.

    Returns:
        NadaArray: A NadaArray with random values of the specified type.
    """
    return NadaArray.random(dims, nada_type)


def output(arr: NadaArray, party: Party, prefix: str):
    """
    Generate a list of Output objects for each element in the input NadaArray.

    Args:
        arr (NadaArray): The input NadaArray.
        party (Party): The party object.
        prefix (str): The prefix for naming the Output objects.

    Returns:
        list: A list of Output objects.
    """
    return NadaArray.output_array(arr, party, prefix)


def vstack(arr_list: list) -> NadaArray:
    """
    Stack arrays in sequence vertically (row wise).

    Args:
        arr_list (list): A list of NadaArray objects to stack.

    Returns:
        NadaArray: The stacked NadaArray.
    """
    return NadaArray(np.vstack(arr_list))


def hstack(arr_list: list) -> NadaArray:
    """
    Stack arrays in sequence horizontally (column wise).

    Args:
        arr_list (list): A list of NadaArray objects to stack.

    Returns:
        NadaArray: The stacked NadaArray.
    """
    return NadaArray(np.hstack(arr_list))
